/*
*		this program performs raw reads on Xenix disks. the result
*		is written to stdout or a named file.
*		
*		this program was inspired by the article "CPIO: System
*		Independent FILE Transfer" C-USERS JOURNAL 6:5
*		(1988:06-07). the basic concept used here follows this
*		article. the actual code is quite different.
*		
*		the code was developed with Turbo C (Borland). see pp.
*		38-41 of the TURBO C REFERENCE GUIDE (Version 1.0) for a
*		discussion of the most critical portion of the code.
*		
*		no attempt has been made to make this program capable of
*		reading formats other than those from SCO Xenix 286 and SCO
*		Xenix 86. this program may be freely distributed provided
*		no charge is made for the distribution. please communicate
*		improvements to the author.
*		
*		Earl H. Kinmonth
*		History Department
*		University of California, Davis
*		Davis, California 95616
*		916-752-1636/0776
*		ucdked!cck@ucdavis.edu
*/
#include	<stdio.h>
#include	<ctype.h>
#include	<bios.h>
#define		FUBAR	69
#define		ZERO	0
#define		MAXTRY	5
#define		PAUSE
#define		NHED	2		/* number of heads on drive */
#define		NSEC	15		/* max number of sectors for one disk track */
#define		NBLK	1		/* number of sectors for one read */
#define		BLKSIZ	512		/* one block (=sector) in chars */
#define		LWIDTH	64		/* display width for debugging */
#define		NTRK	512		/* arbitrary number */
char		buf[BLKSIZ * NSEC];
char		ls[BLKSIZ];
char		sob[BUFSIZ];
char		seb[BUFSIZ];
char		*wfn;
char		*wr_mode = "wb";
int			nsec = NBLK;
int			nblk = NBLK;
int			status;
int			debug;
int			repor;
int			cmd = 2;		/* read sector command */
int			tries;			/* tries before failing */
int			dri = 0;		/* drive: 0 = default, 1 = a, 2 = b, etc. */
int			hed;			/* which head on drive? */
int			trk;			/* which track to read? */
int			sec;			/* which sector to read? */
int			nsc = 1;		/* how many sectors to read? */
int			blk;			/* how many blocks (=secotrs) read? */
int			ntrk = NTRK;	/* number of tracks (arbitrary) */
/*EXTERN*/
extern int atoi(char *);
extern void hint(FILE *);
extern int putcrap(char *);
extern void exit(int);
extern void sleep(int);
extern int strncmp(char *, char *, int);
extern int cmdopt(int,char **);
main(argc,argv)
int		argc;
char	*argv[];
{
	auto	FILE	*stdrpt = stderr;
/**/
	cmdopt(argc,argv);
/**/
	if(debug)	stdrpt = stdout;
/**/
	setbuf(stderr,seb);
	setbuf(stdout,sob);
/**/
	fprintf(stderr,"(c) Earl H. Kinmonth (%s)\n",__DATE__);
	fprintf(stderr,"NON COMMERCIAL USE AND REPRODUCTION ALLOWED.\n");
/**/
	while(1)	{
	fprintf(stderr,"carriage return to read next disk\n");
	fprintf(stderr,"quit (followed by carriage return) to exit\n");
	fprintf(stderr,"     \007insert xenix disk\r");
	fflush(stderr);
	gets(ls);
	if(strncmp(ls,"quit",4) == 0)	break;
	blk = 0;
	for(trk = 0; blk < nblk && trk < ntrk; trk++)		{
		for(hed = 0; blk < nblk && hed < NHED; hed++)	{
			for(sec = 1; blk < nblk && sec <= nsec; sec += nsc)	 {
				fprintf(stdrpt,"blk %4d ",blk);
#ifdef	DEBUG
				fprintf(stdrpt,"sec %4d ",sec);
				fprintf(stdrpt,"nsc %4d ",nsc);
				fprintf(stdrpt,"hed %4d ",hed);
				fprintf(stdrpt,"trk %4d ",trk);
#endif
				if(repor)	fprintf(stdrpt,"\r");
				else		fprintf(stdrpt,"\n");
				fflush(stdrpt);
				/* memset(buf,BLKSIZ,0); */
/*
	1.2 meg disks sometimes require several tries to start reading
*/
				for(tries = 0; tries < MAXTRY; tries++)	{
#ifdef	DEBUG
					fprintf(stderr,"biodisk(%u,%u,%u,%u,%u,%u,%u)\n",
						cmd,dri,hed,trk,sec,nsc,buf);
#endif
					status = biosdisk(cmd,dri,hed,trk,sec,nsc,buf);
					if(status == 0)	break;
					sleep(1);
					}
				if(status)	{
					fprintf(stderr,"\nread error: %07o\n",status);
					exit(69);
					}
				if(debug)	putcrap(buf);
				else		fwrite(buf,sizeof(char),BLKSIZ * nsc,stdout);
				blk += nsc;
			}
		}
	}
	fprintf(stdrpt,"\nblk(s) %4d\n",blk);
	fflush(stdrpt);
	}
	exit(0);
}
putcrap(buf)
char	buf[];
{
	auto	int		ncx = 0;
	auto	int		ix;
	for(ix = 0; ix < BLKSIZ; ix++)	{
		if(isprint(buf[ix]))	putchar(buf[ix]);
		else					putchar('.');
		ncx += 1;
		if(ncx % LWIDTH == 0)	{
			putchar('\n');
			ncx = 0;
			}
		}
		putchar('\n');
}
cmdopt(argc,argv)
int		argc;
char	*argv[];
{
	auto	char	*px;
	auto	int		ix;
	for(ix = 1; ix < argc; ix++)	{
		for(px = argv[ix]; *px != 0; px++)	{
			if(isalpha(*px))	switch(*px)	{
				case	'H':
				case	'h':	hint(stdout);
					break;
				case	'R':
					repor = atoi(px + 1);
					repor = repor ? repor: 1;
					break;
				case	'D':
					debug = atoi(px + 1);
					debug = debug ? debug: 1;
					break;
				case	'a':
					wr_mode[0] = 'a';
					break;
				case	'o':
					if(!(freopen(wfn = argv[++ix],wr_mode,stdout)))	{
						perror(argv[ix]);
						exit(69);
						}
					repor = 1;
					break;
				case	'b':
					nblk = atoi(px + 1);
					break;
				case	'n':
					nsc = atoi(px + 1);
					break;
				case	's':
					nsec = atoi(px + 1);
					break;
				case	't':
					ntrk = atoi(px + 1);
					break;
				case	'A':
					nblk = 720; nsec = 9; nsc = 9; break;
				case	'B':
					nblk = 1440; nsec = 9; nsc = 9; break;
				case	'C':
					nblk = 2400; nsec = 15; nsc = 15; break;
				default:		
					fprintf(stderr,"bad option %c in %s\n",*px,argv[ix]);
					exit(69);
				}
			}
		}
}
static	char	*hx[] = {
	"==========",
	"readisk [-options]",
	"*",
	"*      performs raw disk reads on drive a:",
	"*",
	"==========",
	"-A     .36 meg",
	"-B     .72 meg",
	"-C     1.2 meg",
	"-D     debug",
	"-R     report",
	"-a     append when -o name is used",
	"-bNN   number of blocks to read",
	"-nNN   sectors to read in one disk access",
	"-o     write to the following named file",
	"-sNN   number of sectors",
	"-tNN   number of tracks",
	"==========",
	"EXAMPLES",
	"*      make a 720 tar floppy under Xenix",
	"*          tar 4cv .",
	"*      read this with tar (pd) under MSDOS",
	"*          readisk -B -o x.tar",
	"*          tar -xvf x.tar",
	"==========",
	"NOTES",
	"==========",
	"*      pipes offer no advantage under MSDOS",
	"*      redirection (> file) may cause CR/LF problems",
	"*      blocks and sectors are each 512 characters",
	"==========",
	"BUGS",
	"==========",
	"*      slow",
	"==========",
	NULL,
	}	;
void
hint(fp)
FILE	*fp;
{
	auto	char	**hz = hx;
	auto	int		lc = 0;
	auto	int		sc = 22;
	fputs("\f\n",stdout);
	fflush(stdout);
	while(*hz)	{
		fputs(*hz++,fp);
		fputs("\n",fp);
		if(lc++ && (lc % sc == 0))	{
			fputs("<CR for more>\r",stdout);
			while(getc(stdin) != '\n')	;
			fflush(stdout);
			}
		}
	if(fp == stderr)	exit(FUBAR);
	else				exit(ZERO);
}
